# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import unittest

from threatexchange.signal_type import (
    md5,
    raw_text,
    trend_query,
    url_md5,
    url,
)
from threatexchange.signal_type.pdq import signal
from threatexchange.signal_type.signal_base import TextHasher


class SignalTypeHashTest(unittest.TestCase):
    """
    Sanity check for signal type hashing methods.
    """

    # TODO - maybe make a metaclass for this to automatically detect?
    SIGNAL_TYPES_TO_TEST = [
        md5.VideoMD5Signal,
        signal.PdqSignal,
        raw_text.RawTextSignal,
        trend_query.TrendQuerySignal,
        url_md5.UrlMD5Signal,
        url.URLSignal,
    ]

    def test_signal_names_unique(self):
        seen = {}
        for s in self.SIGNAL_TYPES_TO_TEST:
            name = s.get_name()
            assert (
                name not in seen
            ), f"Two signal types share the same name: {s!r} and {seen[name]!r}"

    def test_signal_types_have_content(self):
        for s in self.SIGNAL_TYPES_TO_TEST:
            assert s.get_content_types(), "{s!r} has no content types"

    def test_str_hashers_have_impl(self):
        text_hashers = [
            s for s in self.SIGNAL_TYPES_TO_TEST if isinstance(s, TextHasher)
        ]
        for s in text_hashers:
            assert s.hash_from_str(
                "test string"
            ), "{s!r} produced no output from hasher"



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import unittest
import pickle
import typing as t
import functools

from threatexchange.signal_type.index import (
    SignalSimilarityInfoWithIntDistance,
)
from threatexchange.signal_type.pdq.pdq_index import PDQIndex, PDQIndexMatch

test_entries = [
    (
        "0000000000000000000000000000000000000000000000000000000000000000",
        dict(
            {
                "hash_type": "pdq",
                "system_id": 9,
            }
        ),
    ),
    (
        "000000000000000000000000000000000000000000000000000000000000ffff",
        dict(
            {
                "hash_type": "pdq",
                "system_id": 8,
            }
        ),
    ),
    (
        "0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f",
        dict(
            {
                "hash_type": "pdq",
                "system_id": 7,
            }
        ),
    ),
    (
        "f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0",
        dict(
            {
                "hash_type": "pdq",
                "system_id": 6,
            }
        ),
    ),
    (
        "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
        dict(
            {
                "hash_type": "pdq",
                "system_id": 5,
            }
        ),
    ),
]


class TestPDQIndex(unittest.TestCase):
    def setUp(self):
        self.index = PDQIndex.build(test_entries)

    def assertEqualPDQIndexMatchResults(
        self, result: t.List[PDQIndexMatch], expected: t.List[PDQIndexMatch]
    ):
        self.assertEqual(
            len(result), len(expected), "search results not of expected length"
        )

        accum_type = t.Dict[int, t.Set[int]]

        # Between python 3.8.6 and 3.8.11, something caused the order of results
        # from the index to change. This was noticed for items which had the
        # same distance. To allow for this, we convert result and expected
        # arrays from
        #   [PDQIndexMatch, PDQIndexMatch] to { distance: <set of PDQIndexMatch.metadata hash> }
        # This allows you to compare [PDQIndexMatch A, PDQIndexMatch B] with
        # [PDQIndexMatch B, PDQIndexMatch A] as long as A.distance == B.distance.
        def quality_indexed_dict_reducer(
            acc: accum_type, item: PDQIndexMatch
        ) -> accum_type:
            acc[item.similarity_info.distance] = acc.get(
                item.similarity_info.distance, set()
            )
            # Instead of storing the unhashable item.metadata dict, store its
            # hash so we can compare using self.assertSetEqual
            acc[item.similarity_info.distance].add(hash(frozenset(item.metadata)))
            return acc

        # Convert results to distance -> set of metadata map
        distance_to_result_items_map: accum_type = functools.reduce(
            quality_indexed_dict_reducer, result, {}
        )

        # Convert expected to distance -> set of metadata map
        distance_to_expected_items_map: accum_type = functools.reduce(
            quality_indexed_dict_reducer, expected, {}
        )

        assert len(distance_to_expected_items_map) == len(
            distance_to_result_items_map
        ), "Unequal number of items in expected and results."

        for distance, result_items in distance_to_result_items_map.items():
            assert (
                distance in distance_to_expected_items_map
            ), f"Unexpected distance {distance} found"
            self.assertSetEqual(result_items, distance_to_expected_items_map[distance])

    def test_search_index_for_matches(self):
        entry_hash = test_entries[1][0]
        result = self.index.query(entry_hash)
        self.assertEqualPDQIndexMatchResults(
            result,
            [
                PDQIndexMatch(
                    SignalSimilarityInfoWithIntDistance(0), test_entries[1][1]
                ),
                PDQIndexMatch(
                    SignalSimilarityInfoWithIntDistance(16), test_entries[0][1]
                ),
            ],
        )

    def test_search_index_with_no_match(self):
        query_hash = "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
        result = self.index.query(query_hash)
        self.assertEqualPDQIndexMatchResults(result, [])

    def test_supports_pickling(self):
        pickled_data = pickle.dumps(self.index)
        assert pickled_data != None, "index does not support pickling to a data stream"

        reconstructed_index = pickle.loads(pickled_data)
        assert (
            reconstructed_index != None
        ), "index does not support unpickling from data stream"
        assert (
            reconstructed_index.index.faiss_index != self.index.index.faiss_index
        ), "unpickling should create it's own faiss index in memory"

        query = test_entries[0][0]
        result = reconstructed_index.query(query)
        self.assertEqualPDQIndexMatchResults(
            result,
            [
                PDQIndexMatch(
                    SignalSimilarityInfoWithIntDistance(0), test_entries[1][1]
                ),
                PDQIndexMatch(
                    SignalSimilarityInfoWithIntDistance(16), test_entries[0][1]
                ),
            ],
        )



import tempfile
from threatexchange.cli.tests.e2e_test_helper import ThreatExchangeCLIE2eTest
from threatexchange.signal_type.md5 import VideoMD5Signal


class MatchCommandTest(ThreatExchangeCLIE2eTest):

    COMMON_CALL_ARGS = ("match",)

    def test_file_noexst(self):
        self.assert_cli_usage_error(("text", "doesnt_exist.txt"))

    def test_match_file(self):
        with tempfile.NamedTemporaryFile() as fp:
            # Empty file
            self.assert_cli_output(
                ("video", fp.name), "video_md5 - (Sample Signals) INVESTIGATION_SEED"
            )

    def test_hash(self):
        hash = VideoMD5Signal.get_examples()[0]
        self.assert_cli_output(
            ("-H", "video", "--", hash),
            "video_md5 - (Sample Signals) INVESTIGATION_SEED",
        )

    def test_invalid_hash(self):
        not_hash = "this is not an md5"
        self.assert_cli_usage_error(
            ("-H", "video", "--", not_hash),
            f"{not_hash!r} from .* is not a valid hash for video_md5",
        )
